"use strict";
/**
 * Copyright (c) 2019-2020 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author Áron Samuel Kovács <aron.kovacs@mail.muni.cz>
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.CutawayPass = exports.CutawayParams = void 0;
var tslib_1 = require("tslib");
/**
 * Adaptive Cutaways for Comprehensible Rendering of Polygonal Scenes by Michael Burns and Adam Finkelstein
 * https://gfx.cs.princeton.edu/pubs/Burns_2008_ACF/adaptive_cutaways.pdf
 */
var util_1 = require("../../mol-gl/compute/util");
var schema_1 = require("../../mol-gl/renderable/schema");
var shader_code_1 = require("../../mol-gl/shader-code");
var render_item_1 = require("../../mol-gl/webgl/render-item");
var renderable_1 = require("../../mol-gl/renderable");
var linear_algebra_1 = require("../../mol-math/linear-algebra");
var mol_util_1 = require("../../mol-util");
var quad_vert_1 = require("../../mol-gl/shader/quad.vert");
var cutaway_jfa_step_frag_1 = require("../../mol-gl/shader/cutaway-jfa-step.frag");
var cutaway_init_frag_1 = require("../../mol-gl/shader/cutaway-init.frag");
var cutaway_copy_to_target_frag_1 = require("../../mol-gl/shader/cutaway-copy-to-target.frag");
var param_definition_1 = require("../../mol-util/param-definition");
var stereo_1 = require("../camera/stereo");
var debug_1 = require("../../mol-util/debug");
exports.CutawayParams = {
    angle: param_definition_1.ParamDefinition.Numeric(45, { min: 0, max: 90, step: 1 }),
    borderSize: param_definition_1.ParamDefinition.Numeric(0.05, { min: 0, max: 0.1, step: 0.01 }),
    slopeOffset: param_definition_1.ParamDefinition.Numeric(0, { min: 0, max: 1, step: 0.01 }),
};
var CutawayInitSchema = tslib_1.__assign(tslib_1.__assign({}, util_1.QuadSchema), { tDepth: schema_1.TextureSpec('texture', 'rgba', 'ubyte', 'nearest'), uTexSize: schema_1.UniformSpec('v2'), uIsOrtho: schema_1.UniformSpec('f'), uNear: schema_1.UniformSpec('f'), uFar: schema_1.UniformSpec('f') });
function getCutawayInitRenderable(ctx, depthTexture) {
    var values = tslib_1.__assign(tslib_1.__assign({}, util_1.QuadValues), { tDepth: mol_util_1.ValueCell.create(depthTexture), uTexSize: mol_util_1.ValueCell.create(linear_algebra_1.Vec2.create(depthTexture.getWidth(), depthTexture.getHeight())), uIsOrtho: mol_util_1.ValueCell.create(0.0), uNear: mol_util_1.ValueCell.create(0.0), uFar: mol_util_1.ValueCell.create(1.0) });
    var schema = tslib_1.__assign({}, CutawayInitSchema);
    var shaderCode = shader_code_1.ShaderCode('cutaway-init', quad_vert_1.quad_vert, cutaway_init_frag_1.cutawayInit_frag);
    var renderItem = render_item_1.createComputeRenderItem(ctx, 'triangles', shaderCode, schema, values);
    return renderable_1.createComputeRenderable(renderItem, values);
}
var CutawayJfaSchema = tslib_1.__assign(tslib_1.__assign({}, util_1.QuadSchema), { tCutaway: schema_1.TextureSpec('texture', 'rgba', 'float', 'nearest'), uTexSize: schema_1.UniformSpec('v2'), uInvProjection: schema_1.UniformSpec('m4'), uViewport: schema_1.UniformSpec('v4'), uStep: schema_1.UniformSpec('v2'), uPMsz: schema_1.UniformSpec('f'), uAngle: schema_1.UniformSpec('f'), uEdgeRegionSize: schema_1.UniformSpec('f'), uSlopeStartOffset: schema_1.UniformSpec('f'), uAspectRatio: schema_1.UniformSpec('v2'), uIsOrtho: schema_1.UniformSpec('f'), uNear: schema_1.UniformSpec('f'), uFar: schema_1.UniformSpec('f') });
function getCutawayJfaRenderable(ctx, cutawayTexture) {
    var values = tslib_1.__assign(tslib_1.__assign({}, util_1.QuadValues), { tCutaway: mol_util_1.ValueCell.create(cutawayTexture), uTexSize: mol_util_1.ValueCell.create(linear_algebra_1.Vec2.create(cutawayTexture.getWidth(), cutawayTexture.getHeight())), uInvProjection: mol_util_1.ValueCell.create(linear_algebra_1.Mat4.identity()), uViewport: mol_util_1.ValueCell.create(linear_algebra_1.Vec4.create(0.0, 0.0, 1.0, 1.0)), uStep: mol_util_1.ValueCell.create(linear_algebra_1.Vec2.create(0, 0)), uPMsz: mol_util_1.ValueCell.create(1.0), uAngle: mol_util_1.ValueCell.create(Math.PI / 3.0), uEdgeRegionSize: mol_util_1.ValueCell.create(0.0), uSlopeStartOffset: mol_util_1.ValueCell.create(0.0), uAspectRatio: mol_util_1.ValueCell.create(linear_algebra_1.Vec2.create(1.0, 1.0)), uIsOrtho: mol_util_1.ValueCell.create(0.0), uNear: mol_util_1.ValueCell.create(0.0), uFar: mol_util_1.ValueCell.create(1.0) });
    var schema = tslib_1.__assign({}, CutawayJfaSchema);
    var shaderCode = shader_code_1.ShaderCode('cutaway-jfa', quad_vert_1.quad_vert, cutaway_jfa_step_frag_1.cutawayJfaStep_frag);
    var renderItem = render_item_1.createComputeRenderItem(ctx, 'triangles', shaderCode, schema, values);
    return renderable_1.createComputeRenderable(renderItem, values);
}
var CutawayCopyToTargetSchema = tslib_1.__assign(tslib_1.__assign({}, util_1.QuadSchema), { tDepth: schema_1.TextureSpec('texture', 'rgba', 'float', 'nearest'), uTexSize: schema_1.UniformSpec('v2'), uIsOrtho: schema_1.UniformSpec('f'), uNear: schema_1.UniformSpec('f'), uFar: schema_1.UniformSpec('f') });
function getCutawayCopyToTargetRenderable(ctx, depthTexture) {
    var values = tslib_1.__assign(tslib_1.__assign({}, util_1.QuadValues), { tDepth: mol_util_1.ValueCell.create(depthTexture), uTexSize: mol_util_1.ValueCell.create(linear_algebra_1.Vec2.create(depthTexture.getWidth(), depthTexture.getHeight())), uIsOrtho: mol_util_1.ValueCell.create(0.0), uNear: mol_util_1.ValueCell.create(0.0), uFar: mol_util_1.ValueCell.create(1.0) });
    var schema = tslib_1.__assign({}, CutawayCopyToTargetSchema);
    var shaderCode = shader_code_1.ShaderCode('cutaway-copy-to-target', quad_vert_1.quad_vert, cutaway_copy_to_target_frag_1.cutawayCopyToTarget_frag);
    var renderItem = render_item_1.createComputeRenderItem(ctx, 'triangles', shaderCode, schema, values);
    return renderable_1.createComputeRenderable(renderItem, values);
}
var CutawayPass = /** @class */ (function () {
    function CutawayPass(webgl, width, height) {
        this.webgl = webgl;
        CutawayPass._supported = CutawayPass.isSupported(webgl);
        if (!CutawayPass.supported) {
            return;
        }
        this._target = webgl.createRenderTarget(width, height, true, 'uint8', 'nearest');
        this.jfaATarget = webgl.createRenderTarget(width, height, false, 'float32', 'nearest');
        this.jfaBTarget = webgl.createRenderTarget(width, height, false, 'float32', 'nearest');
        this.jfaARenderable = getCutawayJfaRenderable(webgl, this.jfaATarget.texture);
        this.jfaBRenderable = getCutawayJfaRenderable(webgl, this.jfaBTarget.texture);
        this.initRenderable = getCutawayInitRenderable(webgl, this._target.texture);
        this.copyToTargetRenderable = getCutawayCopyToTargetRenderable(webgl, this.jfaATarget.texture);
    }
    Object.defineProperty(CutawayPass, "supported", {
        get: function () {
            return this._supported;
        },
        enumerable: false,
        configurable: true
    });
    Object.defineProperty(CutawayPass.prototype, "target", {
        get: function () {
            return CutawayPass.supported ? this._target.texture : null;
        },
        enumerable: false,
        configurable: true
    });
    CutawayPass.isSupported = function (webgl) {
        var _a = webgl.extensions, textureFloat = _a.textureFloat, colorBufferFloat = _a.colorBufferFloat;
        if (!textureFloat || !colorBufferFloat) {
            if (debug_1.isDebugMode) {
                var missing = [];
                if (!textureFloat)
                    missing.push('textureFloat');
                if (!colorBufferFloat)
                    missing.push('colorBufferFloat');
                console.log("Missing \"" + missing.join('", "') + "\" extensions required for \"cutaway\"");
            }
            return false;
        }
        else {
            return true;
        }
    };
    CutawayPass.prototype.setRenderDepthState = function (renderer, camera) {
        var _a = this.webgl, gl = _a.gl, state = _a.state;
        var _b = camera.viewport, x = _b.x, y = _b.y, width = _b.width, height = _b.height;
        renderer.setDrawingBufferSize(this._target.getWidth(), this._target.getHeight());
        renderer.setViewport(x, y, width, height);
        renderer.update(camera);
        gl.viewport(x, y, width, height);
        gl.scissor(x, y, width, height);
        state.enable(gl.SCISSOR_TEST);
        state.enable(gl.DEPTH_TEST);
        state.colorMask(true, true, true, true);
        state.clearColor(1, 1, 1, 1);
        state.depthMask(true);
        gl.clear(gl.COLOR_BUFFER_BIT | gl.DEPTH_BUFFER_BIT);
    };
    CutawayPass.prototype.setViewport = function (viewport) {
        mol_util_1.ValueCell.update(this.jfaARenderable.values.uViewport, linear_algebra_1.Vec4.copy(this.jfaARenderable.values.uViewport.ref.value, viewport));
        mol_util_1.ValueCell.update(this.jfaBRenderable.values.uViewport, linear_algebra_1.Vec4.copy(this.jfaBRenderable.values.uViewport.ref.value, viewport));
    };
    CutawayPass.prototype.setJfaState = function (camera, scene, props) {
        var _a = this.webgl, gl = _a.gl, state = _a.state;
        var orthographic = camera.state.mode === 'orthographic' ? 1 : 0;
        state.enable(gl.SCISSOR_TEST);
        state.disable(gl.BLEND);
        state.disable(gl.DEPTH_TEST);
        state.depthMask(false);
        var PMsz = orthographic ? 0 : (camera.near + camera.far) / (camera.near - camera.far);
        var invProjection = linear_algebra_1.Mat4.identity();
        linear_algebra_1.Mat4.invert(invProjection, camera.projection);
        mol_util_1.ValueCell.updateIfChanged(this.jfaARenderable.values.uInvProjection, invProjection);
        mol_util_1.ValueCell.updateIfChanged(this.jfaBRenderable.values.uInvProjection, invProjection);
        mol_util_1.ValueCell.updateIfChanged(this.jfaARenderable.values.uPMsz, PMsz);
        mol_util_1.ValueCell.updateIfChanged(this.jfaBRenderable.values.uPMsz, PMsz);
        mol_util_1.ValueCell.updateIfChanged(this.initRenderable.values.uIsOrtho, orthographic);
        mol_util_1.ValueCell.updateIfChanged(this.jfaARenderable.values.uIsOrtho, orthographic);
        mol_util_1.ValueCell.updateIfChanged(this.jfaBRenderable.values.uIsOrtho, orthographic);
        mol_util_1.ValueCell.updateIfChanged(this.copyToTargetRenderable.values.uIsOrtho, orthographic);
        mol_util_1.ValueCell.updateIfChanged(this.initRenderable.values.uNear, camera.near);
        mol_util_1.ValueCell.updateIfChanged(this.jfaARenderable.values.uNear, camera.near);
        mol_util_1.ValueCell.updateIfChanged(this.jfaBRenderable.values.uNear, camera.near);
        mol_util_1.ValueCell.updateIfChanged(this.copyToTargetRenderable.values.uNear, camera.near);
        mol_util_1.ValueCell.updateIfChanged(this.initRenderable.values.uFar, camera.far);
        mol_util_1.ValueCell.updateIfChanged(this.jfaARenderable.values.uFar, camera.far);
        mol_util_1.ValueCell.updateIfChanged(this.jfaBRenderable.values.uFar, camera.far);
        mol_util_1.ValueCell.updateIfChanged(this.copyToTargetRenderable.values.uFar, camera.far);
        var angle = Math.max(props.angle * Math.PI / 180.0, 0.001);
        mol_util_1.ValueCell.updateIfChanged(this.jfaARenderable.values.uAngle, angle);
        mol_util_1.ValueCell.updateIfChanged(this.jfaBRenderable.values.uAngle, angle);
        mol_util_1.ValueCell.updateIfChanged(this.jfaARenderable.values.uEdgeRegionSize, props.borderSize);
        mol_util_1.ValueCell.updateIfChanged(this.jfaBRenderable.values.uEdgeRegionSize, props.borderSize);
        var slopeStartOffset = scene.boundingSphereVisible.radius * props.slopeOffset;
        mol_util_1.ValueCell.updateIfChanged(this.jfaARenderable.values.uSlopeStartOffset, slopeStartOffset);
        mol_util_1.ValueCell.updateIfChanged(this.jfaBRenderable.values.uSlopeStartOffset, slopeStartOffset);
        var aspectRatio = this.getScalingRatio(camera, scene);
        mol_util_1.ValueCell.update(this.jfaARenderable.values.uAspectRatio, linear_algebra_1.Vec2.copy(this.jfaARenderable.values.uAspectRatio.ref.value, aspectRatio));
        mol_util_1.ValueCell.update(this.jfaBRenderable.values.uAspectRatio, linear_algebra_1.Vec2.copy(this.jfaBRenderable.values.uAspectRatio.ref.value, aspectRatio));
    };
    CutawayPass.prototype.getScalingRatio = function (camera, scene) {
        var distanceToCamera = linear_algebra_1.Vec3.distance(camera.state.position, scene.boundingSphereVisible.center);
        var center = linear_algebra_1.Vec4.create(0, 0, distanceToCamera, 1);
        var right = linear_algebra_1.Vec4.create(scene.boundingSphereVisible.radius, 0, distanceToCamera, 1);
        var top = linear_algebra_1.Vec4.create(0, scene.boundingSphereVisible.radius, distanceToCamera, 1);
        // project
        linear_algebra_1.Vec4.transformMat4(center, center, camera.projection);
        linear_algebra_1.Vec4.transformMat4(right, right, camera.projection);
        linear_algebra_1.Vec4.transformMat4(top, top, camera.projection);
        // scale only the relevant coordinates
        center[0] = center[0] / center[3];
        center[1] = center[1] / center[3];
        right[0] = right[0] / right[3];
        top[1] = top[1] / top[3];
        // distances between points
        var ratio = linear_algebra_1.Vec2.create(Math.abs(right[0] - center[0]), Math.abs(top[1] - center[1]));
        // NDC -> screen space
        linear_algebra_1.Vec2.scale(ratio, ratio, 0.5);
        // and we want to scale by the inverse
        linear_algebra_1.Vec2.inverse(ratio, ratio);
        linear_algebra_1.Vec2.set(ratio, 1, 1);
        return ratio;
    };
    CutawayPass.prototype.renderDepth = function (renderer, camera, scene, props) {
        this._target.bind();
        this.setRenderDepthState(renderer, camera);
        var primitives = scene.primitives.renderables.concat();
        var ligand = primitives.filter(function (r) { return r.values.cutaway.ref.value; });
        var groupLigand = {
            renderables: ligand,
            direction: scene.primitives.direction,
            position: scene.primitives.position,
            up: scene.primitives.up,
            view: scene.primitives.view
        };
        renderer.renderDepth(groupLigand, camera, null, null);
    };
    CutawayPass.prototype.renderJfa = function (camera, scene, props) {
        this.setJfaState(camera, scene, props);
        // initialize the A JFA buffer
        this.jfaATarget.bind();
        this.initRenderable.render();
        // JFA
        var width = this.jfaATarget.getWidth();
        var height = this.jfaATarget.getHeight();
        var stepNormalized = linear_algebra_1.Vec2();
        var readingA = true;
        var stepPixels = Math.ceil(Math.max(width, height) * 0.5);
        while (true) {
            linear_algebra_1.Vec2.set(stepNormalized, stepPixels / width, stepPixels / height);
            if (readingA) {
                this.jfaBTarget.bind();
                mol_util_1.ValueCell.update(this.jfaARenderable.values.uStep, stepNormalized);
                this.jfaARenderable.render();
            }
            else {
                this.jfaATarget.bind();
                mol_util_1.ValueCell.update(this.jfaBRenderable.values.uStep, stepNormalized);
                this.jfaBRenderable.render();
            }
            readingA = !readingA;
            if (stepPixels <= 1) {
                break;
            }
            stepPixels = Math.ceil(stepPixels / 2);
        }
        // copy from the final JFA buffer
        var finalJfaTexture = readingA ? this.jfaATarget.texture : this.jfaBTarget.texture;
        if (this.copyToTargetRenderable.values.tDepth.ref.value !== finalJfaTexture) {
            mol_util_1.ValueCell.update(this.copyToTargetRenderable.values.tDepth, finalJfaTexture);
            this.copyToTargetRenderable.update();
        }
        this._target.bind();
        this.copyToTargetRenderable.render();
    };
    CutawayPass.shouldRun = function (scene) {
        for (var i = 0; i < scene.primitives.renderables.length; i++) {
            if (scene.primitives.renderables[i].values.cutaway.ref.value) {
                return true;
            }
        }
        return false;
    };
    CutawayPass.prototype.render = function (renderer, camera, scene, props) {
        if (!CutawayPass.supported) {
            return;
        }
        if (!CutawayPass.shouldRun(scene)) {
            this._target.bind();
            if (stereo_1.StereoCamera.is(camera)) {
                this.setRenderDepthState(renderer, camera.left);
                this.setRenderDepthState(renderer, camera.right);
            }
            else {
                this.setRenderDepthState(renderer, camera);
            }
            return;
        }
        if (stereo_1.StereoCamera.is(camera)) {
            this.setViewport(linear_algebra_1.Vec4.create(0, 0, 0.5, 1));
            this.renderDepth(renderer, camera.left, scene, props);
            this.renderJfa(camera.left, scene, props);
            this.setViewport(linear_algebra_1.Vec4.create(0.5, 0, 0.5, 1));
            this.renderDepth(renderer, camera.right, scene, props);
            this.renderJfa(camera.right, scene, props);
        }
        else {
            this.setViewport(linear_algebra_1.Vec4.create(0, 0, 1, 1));
            this.renderDepth(renderer, camera, scene, props);
            this.renderJfa(camera, scene, props);
        }
    };
    CutawayPass.prototype.setSize = function (width, height) {
        var _a = this.jfaARenderable.values.uTexSize.ref.value, w = _a[0], h = _a[1];
        if (width !== w || height !== h) {
            this._target.setSize(width, height);
            this.jfaATarget.setSize(width, height);
            this.jfaBTarget.setSize(width, height);
            mol_util_1.ValueCell.update(this.initRenderable.values.uTexSize, linear_algebra_1.Vec2.set(this.initRenderable.values.uTexSize.ref.value, width, height));
            mol_util_1.ValueCell.update(this.jfaARenderable.values.uTexSize, linear_algebra_1.Vec2.set(this.jfaARenderable.values.uTexSize.ref.value, width, height));
            mol_util_1.ValueCell.update(this.jfaBRenderable.values.uTexSize, linear_algebra_1.Vec2.set(this.jfaBRenderable.values.uTexSize.ref.value, width, height));
            mol_util_1.ValueCell.update(this.copyToTargetRenderable.values.uTexSize, linear_algebra_1.Vec2.set(this.copyToTargetRenderable.values.uTexSize.ref.value, width, height));
        }
    };
    CutawayPass._supported = false;
    return CutawayPass;
}());
exports.CutawayPass = CutawayPass;
//# sourceMappingURL=cutaway.js.map