/**
 * Copyright (c) 2021 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 */
import { __assign } from "tslib";
import { BehaviorSubject } from 'rxjs';
import { Canvas3D, Canvas3DContext, Canvas3DParams, DefaultCanvas3DParams } from '../../mol-canvas3d/canvas3d';
import { Passes } from '../../mol-canvas3d/passes/passes';
import { resizeCanvas } from '../../mol-canvas3d/util';
import { createContext, getGLContext } from '../../mol-gl/webgl/context';
import { Vec3 } from '../../mol-math/linear-algebra';
import { Representation } from '../../mol-repr/representation';
import { isDebugMode } from '../../mol-util/debug';
import { InputObserver } from '../../mol-util/input/input-observer';
import { now } from '../../mol-util/now';
import { ParamDefinition as PD } from '../../mol-util/param-definition';
import { RxEventHelper } from '../../mol-util/rx-event-helper';
var MultiCanvas3D = /** @class */ (function () {
    function MultiCanvas3D(builder) {
        var _this = this;
        this.builder = builder;
        this.ev = RxEventHelper.create();
        this.interaction = {
            hover: this.ev(),
            drag: this.ev(),
            click: this.ev(),
        };
        this.grid = [];
        this.currentCell = -1;
        this.refOrientation = {
            position: Vec3.create(0, 0, 100),
            up: Vec3.create(0, 1, 0),
            target: Vec3.create(0, 0, 0),
        };
        this.cameraSyncState = false;
        this.contextLost = new BehaviorSubject(0);
        this.handleWebglContextLost = function (e) {
            _this.webgl.setContextLost();
            e.preventDefault();
            if (isDebugMode)
                console.log('context lost');
            _this.contextLost.next(now());
        };
        this.handlewWebglContextRestored = function () {
            if (!_this.webgl.isContextLost)
                return;
            _this.webgl.handleContextRestored(function () {
                _this.passes.draw.reset();
            });
            if (isDebugMode)
                console.log('context restored');
        };
        this.container = document.createElement('div');
        Object.assign(this.container.style, {
            position: 'absolute',
            width: '100%',
            height: '100%',
        });
        builder.parent.appendChild(this.container);
        this.canvas = document.createElement('canvas');
        this.container.appendChild(this.canvas);
        resizeCanvas(this.canvas, this.container);
        this.input = InputObserver.fromElement(this.canvas, {
            mask: function (x, y) {
                for (var i = 0, il = builder.elementCount; i < il; ++i) {
                    var element = builder.getElement(i);
                    var r = element.getBoundingClientRect();
                    if (x >= r.left && x <= r.right && y >= r.top && y <= r.bottom)
                        return true;
                }
                return false;
            }
        });
        var gl = getGLContext(this.canvas, {
            /** true by default to avoid issues with Safari (Jan 2021) */
            antialias: true,
            /** true to support multiple Canvas3D objects with a single context */
            preserveDrawingBuffer: true,
            alpha: true,
            depth: true,
            premultipliedAlpha: true,
        });
        if (gl === null)
            throw new Error('Could not create a WebGL rendering context');
        this.webgl = createContext(gl);
        this.passes = new Passes(this.webgl, { pickScale: 1, enableWboit: true });
        builder.parent.addEventListener('scroll', function () { return _this.draw(); });
        this.input.resize.subscribe(function () { return _this.handleResize(); });
        this.input.drag.subscribe(function (_a) {
            var pageX = _a.pageX, pageY = _a.pageY, isStart = _a.isStart;
            if (isStart)
                _this.updateCurrentCell(pageX, pageY);
            if (_this.cameraSyncState) {
                // schedule for next frame when current cell is updated
                requestAnimationFrame(function () { return _this.syncOrientation(); });
            }
        });
        this.input.wheel.subscribe(function (_a) {
            var pageX = _a.pageX, pageY = _a.pageY;
            _this.updateCurrentCell(pageX, pageY);
            if (_this.cameraSyncState) {
                // schedule for next frame when current cell is updated
                requestAnimationFrame(function () { return _this.syncOrientation(); });
            }
        });
        this.input.interactionEnd.subscribe(function () {
            _this.currentCell = -1;
        });
    }
    MultiCanvas3D.prototype.updateCurrentCell = function (x, y) {
        this.currentCell = -1;
        for (var i = 0, il = this.builder.elementCount; i < il; ++i) {
            var r = this.builder.getElement(i).getBoundingClientRect();
            if (x >= r.left && x <= r.right && y <= r.bottom && y >= r.top) {
                this.currentCell = i;
                break;
            }
        }
    };
    Object.defineProperty(MultiCanvas3D.prototype, "cameraSync", {
        get: function () { return this.cameraSyncState; },
        set: function (state) {
            if (state === this.cameraSyncState)
                return;
            if (state)
                this.syncOrientation();
            else
                this.resetCamera();
            this.cameraSyncState = state;
        },
        enumerable: false,
        configurable: true
    });
    MultiCanvas3D.prototype.resetCamera = function () {
        var _a;
        for (var i = 0, il = this.builder.elementCount; i < il; ++i) {
            (_a = this.grid[i]) === null || _a === void 0 ? void 0 : _a.requestCameraReset();
        }
    };
    MultiCanvas3D.prototype.syncOrientation = function () {
        var cc = this.grid[this.currentCell];
        if (cc)
            copyOrientation(this.refOrientation, cc.camera.state);
        for (var i = 0, il = this.builder.elementCount; i < il; ++i) {
            var c = this.grid[i];
            if (c)
                copyOrientation(c.camera.state, this.refOrientation);
        }
    };
    MultiCanvas3D.prototype.draw = function () {
        var _a;
        // move the canvas to the current scroll position
        var _b = this.builder.parent, scrollLeft = _b.scrollLeft, scrollTop = _b.scrollTop;
        this.container.style.transform = "translate(" + scrollLeft + "px, " + scrollTop + "px)";
        // clear canvas
        this.webgl.clear(0, 0, 0, 0);
        var _c = this.interaction, click = _c.click, drag = _c.drag, hover = _c.hover;
        var r = this.canvas.getBoundingClientRect();
        for (var i = 0, il = this.builder.elementCount; i < il; ++i) {
            var element = this.builder.getElement(i);
            var rect = element.getBoundingClientRect();
            if (rect.left > r.right || rect.right < r.left ||
                rect.bottom < r.top || rect.top > r.bottom) {
                (_a = this.grid[i]) === null || _a === void 0 ? void 0 : _a.pause();
            }
            else {
                var viewport = {
                    name: 'static-frame',
                    params: {
                        x: rect.left - r.left,
                        y: rect.top - r.top,
                        width: rect.right - rect.left,
                        height: rect.bottom - rect.top
                    }
                };
                var c = this.grid[i];
                if (c === undefined) {
                    var p = this.builder.props(i);
                    c = Canvas3D.create({
                        canvas: this.canvas,
                        webgl: this.webgl,
                        input: this.input,
                        passes: this.passes,
                        attribs: Canvas3DContext.DefaultAttribs,
                        contextLost: this.contextLost,
                        contextRestored: this.webgl.contextRestored,
                        dispose: function () { }
                    }, __assign(__assign({}, PD.merge(Canvas3DParams, DefaultCanvas3DParams, p)), { viewport: viewport }));
                    this.builder.init(i, c);
                    this.grid[i] = c;
                    c.animate();
                    if (this.cameraSyncState) {
                        copyOrientation(c.camera.state, this.refOrientation);
                    }
                    else {
                        c.requestCameraReset();
                    }
                    c.interaction.click.subscribe(function (e) { return click.next(e); });
                    c.interaction.drag.subscribe(function (e) { return drag.next(e); });
                    c.interaction.hover.subscribe(function (e) { return hover.next(e); });
                }
                else {
                    c.setProps({ viewport: viewport });
                    c.animate();
                }
            }
        }
    };
    MultiCanvas3D.prototype.handleResize = function () {
        resizeCanvas(this.canvas, this.container);
        this.passes.updateSize();
        this.draw();
    };
    MultiCanvas3D.prototype.animate = function () {
        for (var _i = 0, _a = this.grid; _i < _a.length; _i++) {
            var c = _a[_i];
            c === null || c === void 0 ? void 0 : c.animate();
        }
    };
    MultiCanvas3D.prototype.pause = function () {
        for (var _i = 0, _a = this.grid; _i < _a.length; _i++) {
            var c = _a[_i];
            c === null || c === void 0 ? void 0 : c.pause();
        }
    };
    MultiCanvas3D.prototype.identify = function (x, y) {
        var r = this.builder.parent.getBoundingClientRect();
        if (x < 0 || x > r.width || y < 0 || y > r.height)
            return;
        for (var i = 0, il = this.builder.elementCount; i < il; ++i) {
            var element = this.builder.getElement(i);
            var rect = element.getBoundingClientRect();
            var c = this.grid[i];
            if (c === undefined ||
                rect.left > r.right || rect.right < r.left ||
                rect.bottom < r.top || rect.top > r.bottom)
                continue;
            var pickingId = c.identify(x, y);
            if (pickingId)
                return pickingId;
        }
    };
    MultiCanvas3D.prototype.mark = function (loci, action) {
        for (var _i = 0, _a = this.grid; _i < _a.length; _i++) {
            var c = _a[_i];
            c === null || c === void 0 ? void 0 : c.mark(loci, action);
        }
    };
    MultiCanvas3D.prototype.getLoci = function (pickingId) {
        for (var _i = 0, _a = this.grid; _i < _a.length; _i++) {
            var c = _a[_i];
            var loci = c === null || c === void 0 ? void 0 : c.getLoci(pickingId);
            if (loci && !Representation.Loci.isEmpty(loci))
                return loci;
        }
        return Representation.Loci.Empty;
    };
    MultiCanvas3D.prototype.setProps = function (props) {
        for (var _i = 0, _a = this.grid; _i < _a.length; _i++) {
            var c = _a[_i];
            c === null || c === void 0 ? void 0 : c.setProps(props);
        }
    };
    MultiCanvas3D.prototype.dispose = function (options) {
        for (var _i = 0, _a = this.grid; _i < _a.length; _i++) {
            var c = _a[_i];
            c === null || c === void 0 ? void 0 : c.dispose();
        }
        this.input.dispose();
        this.canvas.removeEventListener('webglcontextlost', this.handleWebglContextLost, false);
        this.canvas.removeEventListener('webglcontextrestored', this.handlewWebglContextRestored, false);
        this.webgl.destroy(options);
    };
    return MultiCanvas3D;
}());
export { MultiCanvas3D };
function copyOrientation(a, b) {
    Vec3.copy(a.position, b.position);
    Vec3.copy(a.up, b.up);
    Vec3.copy(a.target, b.target);
}
//# sourceMappingURL=multi-canvas3d.js.map